@@ -14,6 +14,10 @@
    You should have received a copy of the GNU Lesser General Public License
    along with this program; if not, see <http://www.gnu.org/licenses/>.
 */
+
+#ifndef _LIBNFS_PRIVATE_H_
+#define _LIBNFS_PRIVATE_H_
+
 #ifdef HAVE_CONFIG_H
 #include "config.h"  /* HAVE_SOCKADDR_STORAGE ? */
 #endif
@@ -24,6 +28,9 @@
 
 #include "libnfs-zdr.h"
 
+#ifdef __cplusplus
+extern "C" {
+#endif
 
 #if !defined(HAVE_SOCKADDR_STORAGE) && !defined(WIN32)
 /*
@@ -58,6 +65,16 @@ struct rpc_fragment {
 #define RPC_CONTEXT_MAGIC 0xc6e46435
 #define RPC_PARAM_UNDEFINED -1
 
+/*
+ * Queue is singly-linked but we hold on to the tail
+ */
+struct rpc_queue {
+	struct rpc_pdu *head, *tail;
+};
+
+#define HASHES 1024
+#define NFS_RA_TIMEOUT 5
+
 struct rpc_context {
 	uint32_t magic;
 	int fd;
@@ -75,9 +92,9 @@ struct rpc_context {
 	char *encodebuf;
 	int encodebuflen;
 
-	struct rpc_pdu *outqueue;
+	struct rpc_queue outqueue;
 	struct sockaddr_storage udp_src;
-	struct rpc_pdu *waitpdu;
+	struct rpc_queue waitpdu[HASHES];
 
 	uint32_t inpos;
 	uint32_t insize;
@@ -94,11 +111,12 @@ struct rpc_context {
 
 	/* fragment reassembly */
 	struct rpc_fragment *fragments;
-	
+
 	/* parameters passable via URL */
 	int tcp_syncnt;
 	int uid;
 	int gid;
+	uint32_t readahead;
 };
 
 struct rpc_pdu {
@@ -119,6 +137,11 @@ struct rpc_pdu {
 	uint32_t zdr_decode_bufsize;
 };
 
+void rpc_reset_queue(struct rpc_queue *q);
+void rpc_enqueue(struct rpc_queue *q, struct rpc_pdu *pdu);
+void rpc_return_to_queue(struct rpc_queue *q, struct rpc_pdu *pdu);
+unsigned int rpc_hash_xid(uint32_t xid);
+
 struct rpc_pdu *rpc_allocate_pdu(struct rpc_context *rpc, int program, int version, int procedure, rpc_cb cb, void *private_data, zdrproc_t zdr_decode_fn, int zdr_bufsize);
 void rpc_free_pdu(struct rpc_context *rpc, struct rpc_pdu *pdu);
 int rpc_queue_pdu(struct rpc_context *rpc, struct rpc_pdu *pdu);
@@ -126,8 +149,17 @@ int rpc_get_pdu_size(char *buf);
 int rpc_process_pdu(struct rpc_context *rpc, char *buf, int size);
 void rpc_error_all_pdus(struct rpc_context *rpc, char *error);
 
-void rpc_set_error(struct rpc_context *rpc, char *error_string, ...);
-void nfs_set_error(struct nfs_context *nfs, char *error_string, ...);
+void rpc_set_error(struct rpc_context *rpc, char *error_string, ...)
+#ifdef __GNUC__
+ __attribute__((format(printf, 2, 3)))
+#endif
+;
+
+void nfs_set_error(struct nfs_context *nfs, char *error_string, ...)
+#ifdef __GNUC__
+ __attribute__((format(printf, 2, 3)))
+#endif
+;
 
 const char *nfs_get_server(struct nfs_context *nfs);
 const char *nfs_get_export(struct nfs_context *nfs);
@@ -144,8 +176,15 @@ void rpc_unset_autoreconnect(struct rpc_context *rpc);
 void rpc_set_tcp_syncnt(struct rpc_context *rpc, int v);
 void rpc_set_uid(struct rpc_context *rpc, int uid);
 void rpc_set_gid(struct rpc_context *rpc, int gid);
+void rpc_set_readahead(struct rpc_context *rpc, uint32_t v);
 
 int rpc_add_fragment(struct rpc_context *rpc, char *data, uint64_t size);
 void rpc_free_all_fragments(struct rpc_context *rpc);
 
 const struct nfs_fh3 *nfs_get_rootfh(struct nfs_context *nfs);
+
+#ifdef __cplusplus
+}
+#endif
+
+#endif /* !_LIBNFS_PRIVATE_H_ */
