@@ -1,7 +1,10 @@
 /*
    Copyright (C) 2010 by Ronnie Sahlberg <ronniesahlberg@gmail.com>
 
-
+   This program is free software; you can redistribute it and/or modify
+   it under the terms of the GNU Lesser General Public License as published by
+   the Free Software Foundation; either version 2.1 of the License, or
+   (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -49,6 +52,7 @@ struct rpc_context *rpc_init_context(void)
 {
 	struct rpc_context *rpc;
 	static uint32_t salt = 0;
+	unsigned int i;
 
 	rpc = malloc(sizeof(struct rpc_context));
 	if (rpc == NULL) {
@@ -76,17 +80,26 @@ struct rpc_context *rpc_init_context(void)
 	salt += 0x01000000;
 	rpc->fd = -1;
 	rpc->tcp_syncnt = RPC_PARAM_UNDEFINED;
-#ifdef WIN32
+#if defined(WIN32) || defined(ANDROID)
 	rpc->uid = 65534;
 	rpc->gid = 65534;
 #else
 	rpc->uid = getuid();
 	rpc->gid = getgid();
 #endif
+	rpc_reset_queue(&rpc->outqueue);
+	for (i = 0; i < HASHES; i++)
+		rpc_reset_queue(&rpc->waitpdu[i]);
 
 	return rpc;
 }
 
+void rpc_set_readahead(struct rpc_context *rpc, uint32_t v)
+{
+	assert(rpc->magic == RPC_CONTEXT_MAGIC);
+
+	rpc->readahead = v;
+}
 
 struct rpc_context *rpc_init_udp_context(void)
 {
@@ -132,16 +145,18 @@ void rpc_set_gid(struct rpc_context *rpc, int gid) {
 void rpc_set_error(struct rpc_context *rpc, char *error_string, ...)
 {
         va_list ap;
+	char *old_error_string = rpc->error_string;
 
 	assert(rpc->magic == RPC_CONTEXT_MAGIC);
 
-	if (rpc->error_string != NULL) {
-		free(rpc->error_string);
-	}
         va_start(ap, error_string);
 	rpc->error_string = malloc(1024);
 	vsnprintf(rpc->error_string, 1024, error_string, ap);
         va_end(ap);
+
+	if (old_error_string != NULL) {
+		free(old_error_string);
+	}
 }
 
 char *rpc_get_error(struct rpc_context *rpc)
@@ -153,19 +168,27 @@ char *rpc_get_error(struct rpc_context *rpc)
 
 void rpc_error_all_pdus(struct rpc_context *rpc, char *error)
 {
-	struct rpc_pdu *pdu;
+	struct rpc_pdu *pdu, *next;
+	unsigned int i;
 
 	assert(rpc->magic == RPC_CONTEXT_MAGIC);
 
-	while((pdu = rpc->outqueue) != NULL) {
+	while ((pdu = rpc->outqueue.head) != NULL) {
 		pdu->cb(rpc, RPC_STATUS_ERROR, error, pdu->private_data);
-		SLIST_REMOVE(&rpc->outqueue, pdu);
+		rpc->outqueue.head = pdu->next;
 		rpc_free_pdu(rpc, pdu);
 	}
-	while((pdu = rpc->waitpdu) != NULL) {
-		pdu->cb(rpc, RPC_STATUS_ERROR, error, pdu->private_data);
-		SLIST_REMOVE(&rpc->waitpdu, pdu);
-		rpc_free_pdu(rpc, pdu);
+	rpc->outqueue.tail = NULL;
+
+	for (i = 0; i < HASHES; i++) {
+		struct rpc_queue *q = &rpc->waitpdu[i];
+
+		while((pdu = q->head) != NULL) {
+			pdu->cb(rpc, RPC_STATUS_ERROR, error, pdu->private_data);
+			q->head = pdu->next;
+			rpc_free_pdu(rpc, pdu);
+		}
+		q->tail = NULL;
 	}
 }
 
@@ -184,7 +207,7 @@ void rpc_free_all_fragments(struct rpc_context *rpc)
 	while (rpc->fragments != NULL) {
 	      struct rpc_fragment *fragment = rpc->fragments;
 
-	      SLIST_REMOVE(&rpc->fragments, fragment);
+	      rpc->fragments = fragment->next;
 	      rpc_free_fragment(fragment);
 	}
 }
@@ -208,25 +231,31 @@ int rpc_add_fragment(struct rpc_context *rpc, char *data, uint64_t size)
 	}
 
 	memcpy(fragment->data, data, fragment->size);
-	SLIST_ADD_END(&rpc->fragments, fragment);
+	LIBNFS_LIST_ADD_END(&rpc->fragments, fragment);
 	return 0;
 }
 
 void rpc_destroy_context(struct rpc_context *rpc)
 {
 	struct rpc_pdu *pdu;
+	unsigned int i;
 
 	assert(rpc->magic == RPC_CONTEXT_MAGIC);
 
-	while((pdu = rpc->outqueue) != NULL) {
+	while((pdu = rpc->outqueue.head) != NULL) {
 		pdu->cb(rpc, RPC_STATUS_CANCEL, NULL, pdu->private_data);
-		SLIST_REMOVE(&rpc->outqueue, pdu);
+		LIBNFS_LIST_REMOVE(&rpc->outqueue.head, pdu);
 		rpc_free_pdu(rpc, pdu);
 	}
-	while((pdu = rpc->waitpdu) != NULL) {
-		pdu->cb(rpc, RPC_STATUS_CANCEL, NULL, pdu->private_data);
-		SLIST_REMOVE(&rpc->waitpdu, pdu);
-		rpc_free_pdu(rpc, pdu);
+
+	for (i = 0; i < HASHES; i++) {
+		struct rpc_queue *q = &rpc->waitpdu[i];
+
+		while((pdu = q->head) != NULL) {
+			pdu->cb(rpc, RPC_STATUS_CANCEL, NULL, pdu->private_data);
+			LIBNFS_LIST_REMOVE(&q->head, pdu);
+			rpc_free_pdu(rpc, pdu);
+		}
 	}
 
 	rpc_free_all_fragments(rpc);
